from django import forms
from django.contrib.auth import get_user_model
from django.core.exceptions import ValidationError
from django.forms import inlineformset_factory, TextInput, Textarea

from ddm.models.core import ResearchProfile, DonationProject, DonationBlueprint, ProcessingRule, FileUploader

User = get_user_model()


class ProjectCreateForm(forms.ModelForm):
    secret = forms.CharField(min_length=10, max_length=150, required=False, widget=forms.PasswordInput())
    secret_confirm = forms.CharField(min_length=10, max_length=150, required=False, widget=forms.PasswordInput())

    class Meta:
        model = DonationProject
        fields = ['name', 'slug', 'super_secret', 'owner', 'contact_information', 'data_protection_statement']
        widgets = {'owner': forms.HiddenInput()}

    field_order = ['name', 'slug', 'super_secret', 'secret', 'secret_confirm', 'contact_information', 'data_protection_statement']

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if 'initial' in kwargs:
            self.fields['owner'].initial = kwargs['initial']['owner']

    def clean(self):
        super_secret = self.data.get('super_secret', False)
        secret = self.data.get('secret', None)
        secret_confirm = self.data.get('secret_confirm', None)

        if super_secret and secret in ['', None]:
            raise ValidationError('Super secret project needs a secret.')

        if super_secret and (secret != secret_confirm):
            raise ValidationError('Secret and Secret Confirm does not match".')

        super().clean()

    def save(self, commit=True):
        project = super().save(commit=False)
        if project.super_secret:
            project.secret_key = self.data['secret']
        project.save()
        return project


class ResearchProfileConfirmationForm(forms.ModelForm):
    confirmed = forms.BooleanField()

    class Meta:
        model = ResearchProfile
        fields = ['user']
        widgets = {'user': forms.HiddenInput()}

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.expected_user_id = kwargs['initial']['user']
        self.fields['user'].initial = self.expected_user_id
        self.fields['confirmed'].initial = True
        self.fields['confirmed'].widget = forms.HiddenInput()

    def is_valid(self):
        if super().is_valid():
            return self.cleaned_data['confirmed']
        else:
            return False

    def save(self):
        """
        Ensure that user.id is not changed manually (e.g., by altering the HTML).
        """
        if self.cleaned_data['user'].pk != self.expected_user_id:
            raise ValidationError('User is not set as expected.', code='not allowed')
        return super().save()


class BlueprintEditForm(forms.ModelForm):

    class Meta:
        model = DonationBlueprint
        fields = ['name', 'exp_file_format', 'csv_delimiter', 'file_uploader', 'regex_path',
                  'expected_fields']
        widgets = {
            'expected_fields': forms.Textarea(attrs={'rows': 3}),
            'regex_path': forms.Textarea(attrs={'rows': 3}),
        }

    def clean(self):
        related_file_uploader = self.data.get('file_uploader', None)
        regex = self.data.get('regex_path', None)

        if related_file_uploader:
            file_uploader = FileUploader.objects.get(pk=related_file_uploader)
            if file_uploader.upload_type == FileUploader.UploadTypes.ZIP_FILE and regex in ['', None]:
                raise ValidationError(
                    'Donation Blueprints that belong to a ZIP file uploader must define '
                    'a regex pattern.'
                )
        super().clean()


class ProcessingRuleForm(forms.ModelForm):

    class Meta:
        model = ProcessingRule
        fields = ['execution_order', 'name', 'field', 'comparison_operator', 'comparison_value']
        widgets = {
            'field': TextInput(),
            'comparison_value': Textarea(attrs={'cols': 60, 'rows': 2})
        }


ProcessingRuleInlineFormset = inlineformset_factory(
    DonationBlueprint,
    ProcessingRule,
    form=ProcessingRuleForm,
    extra=0
)


class APITokenCreationForm(forms.Form):
    expiration_days = forms.IntegerField(
        initial=30,
        min_value=1,
        max_value=90,
        required=True
    )
    action = forms.CharField(
        max_length=20,
        initial='create',
        widget=forms.HiddenInput()
    )



# Generated by Django 3.2.13 on 2023-01-30 16:33

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('ddm', '0031_auto_20230122_1828'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='questionbase',
            options={'ordering': ['index']},
        ),
        migrations.AlterModelManagers(
            name='matrixquestion',
            managers=[
            ],
        ),
        migrations.AlterModelManagers(
            name='multichoicequestion',
            managers=[
            ],
        ),
        migrations.AlterModelManagers(
            name='openquestion',
            managers=[
            ],
        ),
        migrations.AlterModelManagers(
            name='questionbase',
            managers=[
            ],
        ),
        migrations.AlterModelManagers(
            name='semanticdifferential',
            managers=[
            ],
        ),
        migrations.AlterModelManagers(
            name='singlechoicequestion',
            managers=[
            ],
        ),
        migrations.AlterModelManagers(
            name='transition',
            managers=[
            ],
        ),
        migrations.AlterField(
            model_name='questionbase',
            name='blueprint',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='ddm.donationblueprint'),
        ),
        migrations.AlterField(
            model_name='questionbase',
            name='index',
            field=models.PositiveIntegerField(default=1, verbose_name='Page'),
        ),
    ]



# Generated by Django 3.2.13 on 2023-01-03 13:33

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('ddm', '0021_auto_20221222_0925'),
    ]

    operations = [
        migrations.CreateModel(
            name='FileUploader',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=250)),
                ('index', models.PositiveIntegerField()),
                ('upload_type', models.CharField(choices=[('zip_file', 'Zip File'), ('single_file', 'Single File')], default='single_file', max_length=20, verbose_name='Upload type')),
            ],
        ),
        migrations.RemoveField(
            model_name='blueprintcontainer',
            name='project',
        ),
        migrations.RemoveConstraint(
            model_name='donationinstruction',
            name='unique_index_per_blueprint',
        ),
        migrations.RemoveConstraint(
            model_name='donationinstruction',
            name='unique_index_per_container',
        ),
        migrations.RemoveField(
            model_name='donationblueprint',
            name='blueprint_container',
        ),
        migrations.RemoveField(
            model_name='donationblueprint',
            name='index',
        ),
        migrations.RemoveField(
            model_name='donationinstruction',
            name='blueprint',
        ),
        migrations.RemoveField(
            model_name='donationinstruction',
            name='blueprint_container',
        ),
        migrations.DeleteModel(
            name='BlueprintContainer',
        ),
        migrations.AddField(
            model_name='fileuploader',
            name='project',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='ddm.donationproject'),
        ),
        migrations.AddField(
            model_name='donationblueprint',
            name='file_uploader',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='ddm.fileuploader', verbose_name='Associated File Uploader'),
        ),
        migrations.AddField(
            model_name='donationinstruction',
            name='file_uploader',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to='ddm.fileuploader', verbose_name='Associated File Uploader'),
        ),
        migrations.AddConstraint(
            model_name='donationinstruction',
            constraint=models.UniqueConstraint(fields=('index', 'file_uploader'), name='unique_index_per_file_uploader'),
        ),
    ]



from django.contrib.auth import get_user_model
from django.core.exceptions import ValidationError
from django.test import TestCase, override_settings

from ddm.models.core import DonationProject, ResearchProfile


User = get_user_model()


@override_settings(DDM_SETTINGS={'EMAIL_PERMISSION_CHECK':  r'.*(\.|@)mail\.com$', })
class TestDonationProject(TestCase):
    def setUp(self):
        credentials = {
            'username': 'no_prof', 'password': '123', 'email': 'u@mail.com'
        }
        self.user = User.objects.create_user(**credentials)
        self.user_profile = ResearchProfile.objects.create(user=self.user)

    def test_project_owner_cannot_be_null_on_create(self):
        invalid_project = DonationProject(name='test_project', slug='test')
        self.assertRaises(ValidationError, invalid_project.save)



# Generated by Django 3.2.13 on 2023-01-10 16:24

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ddm', '0025_rename_customtoken_projectaccesstoken'),
    ]

    operations = [
        migrations.AddField(
            model_name='exceptionlogentry',
            name='raised_by',
            field=models.CharField(blank=True, choices=[('server', 'Server'), ('client', 'Client')], max_length=20),
        ),
    ]



import io
import json
import zipfile

from django.core.exceptions import ObjectDoesNotExist
from django.http import HttpResponse
from django.views.decorators.debug import sensitive_variables

from ddm.models.auth import ProjectTokenAuthenticator
from ddm.models.core import (
    DataDonation, DonationProject, QuestionnaireResponse, ResearchProfile,
    Participant, DonationBlueprint
)
from ddm.models.logs import EventLogEntry, ExceptionLogEntry
from ddm.models.serializers import (
    DonationSerializer, ResponseSerializer, ProjectSerializer
)

from rest_framework.views import APIView
from rest_framework.response import Response
from rest_framework import authentication, exceptions, permissions, status


def user_is_allowed_to_download(user, project):
    """
    Returns true if user is owner or collaborator of a project. False otherwise.
    """
    user_profile = ResearchProfile.objects.filter(user=user).first()
    if not user_profile:
        return False
    else:
        if project.owner != user_profile:
            return False
        else:
            return True


class ProjectDataAPI(APIView):
    """
    Download or delete data collected within a project.

    * Token authentication for remote calls.
    * Session authentication for browser access.
    * Only project owners are able to access this view.
    """
    authentication_classes = [ProjectTokenAuthenticator,
                              authentication.SessionAuthentication]
    permission_classes = [permissions.IsAuthenticated]

    def permission_denied(self, request, message=None, code=None):
        """
        If request is not permitted, determine what kind of exception to raise.
        Added EventLog entries.
        """
        if request.authenticators and not request.successful_authenticator:
            self.create_event_log(
                descr=f'Failed {request.method} Attempt',
                msg='Authentication failed.'
            )
            raise exceptions.NotAuthenticated()

        self.create_event_log(
            descr=f'Failed {request.method} Attempt',
            msg='Permission Denied.'
        )
        raise exceptions.PermissionDenied(detail=message, code=code)

    def get_project(self):
        """ Returns project instance. """
        return DonationProject.objects.filter(pk=self.kwargs['pk']).first()

    def create_event_log(self, descr, msg):
        """ Creates an event log entry related to the current project. """
        return EventLogEntry.objects.create(project=self.get_project(),
                                            description=descr,
                                            message=msg)

    @sensitive_variables()
    def get(self, request, format=None, *args, **kwargs):
        """
        Return a zip container that contains a json file which holds the
        data donations and questionnaire responses.
        """
        project = self.get_project()
        if not user_is_allowed_to_download(request.user, project):
            self.create_event_log(
                descr='Forbidden Download Request',
                msg='Request user is not permitted to download the data.'
            )
            msg = 'User does not have access.'
            return Response(status=status.HTTP_403_FORBIDDEN,
                            data={'message': msg})

        # Extract secret from request if project is super secret.
        kwargs = {}
        if project.super_secret:
            secret = None if 'Super-Secret' not in request.headers else request.headers['Super-Secret']
            if secret is None:
                self.create_event_log(
                    descr='Failed Download Attempt',
                    msg='Download requested without supplying secret.'
                )
                msg = 'Incorrect key material.'
                return Response(status=status.HTTP_403_FORBIDDEN,
                                data={'message': msg})
            else:
                kwargs['secret'] = secret

        # Gather project data in dictionary.
        data_donations = DataDonation.objects.filter(project=project)
        q_responses = QuestionnaireResponse.objects.filter(project=project)
        try:
            results = {
                'project': ProjectSerializer(project).data,
                'donations': [DonationSerializer(d, **kwargs).data for d in data_donations],
                'responses': [ResponseSerializer(r, **kwargs).data for r in q_responses],
            }
        except ValueError:
            self.create_event_log(
                descr='Failed Download Attempt',
                msg='Download requested with incorrect secret.'
            )
            msg = 'Incorrect key material.'
            return Response(status=status.HTTP_403_FORBIDDEN,
                            data={'message': msg})

        # Create zip file.
        zip_in_mem = self.create_zip(results)
        response = self.create_zip_response(zip_in_mem)
        self.create_event_log(
            descr='Data Download Successful',
            msg='The project data was downloaded.'
        )
        return response

    @staticmethod
    def create_zip(content):
        """ Creates a zip file in memory. """
        buffer = io.BytesIO()
        with zipfile.ZipFile(buffer, mode='w', compression=zipfile.ZIP_DEFLATED) as zf:
            with zf.open('data.json', 'w') as json_file:
                json_file.write(json.dumps(content, ensure_ascii=False, separators=(',', ':')).encode('utf-8'))
                zf.testzip()
        zip_in_memory = buffer.getvalue()
        buffer.flush()
        return zip_in_memory

    @staticmethod
    def create_zip_response(zip_file):
        """ Creates an HttpResponse object containing the provided zip file. """
        response = HttpResponse(zip_file, content_type='application/zip')
        response['Content-Length'] = len(zip_file)
        response['Content-Disposition'] = 'attachment; filename=zipfile.zip'
        return response

    def delete(self, request, format=None, *args, **kwargs):
        """
        Delete data donations and questionnaire requests associated with project.
        """
        project = self.get_project()
        if not user_is_allowed_to_download(request.user, project):
            self.create_event_log(
                descr='Forbidden Deletion Request',
                msg='Request user is not permitted to delete the project data.'
            )
            msg = 'User does not have access.'
            return Response(status=status.HTTP_403_FORBIDDEN,
                            data={'message': msg})

        # Delete all related objects.
        Participant.objects.filter(project=project).delete()
        n_deleted_donations = DataDonation.objects.filter(project=project).delete()[0]
        n_deleted_responses = QuestionnaireResponse.objects.filter(project=project).delete()[0]

        self.create_event_log(
            descr='Data Deletion Successful',
            msg='The project data was deleted.'
        )

        msg = (f'Deleted {n_deleted_donations} data donations and '
               f'{n_deleted_responses} questionnaire responses.')
        return Response(status=status.HTTP_200_OK, data={'message': msg})


class ParticipantAPI(APIView):
    """
    Delete a participant of an owned project by providing their external ID.

    * Session authentication for browser access.
    * Only project owners are able to access this view.
    """
    authentication_classes = [ProjectTokenAuthenticator,
                              authentication.SessionAuthentication]
    permission_classes = [permissions.IsAuthenticated]

    def permission_denied(self, request, message=None, code=None):
        """
        If request is not permitted, determine what kind of exception to raise.
        Added EventLog entries.
        """
        if request.authenticators and not request.successful_authenticator:
            self.create_event_log(
                descr='Failed participant deletion Attempt',
                msg='Authentication failed.'
            )
            raise exceptions.NotAuthenticated()

        self.create_event_log(
            descr='Failed participant deletion Attempt',
            msg='Permission Denied.'
        )
        raise exceptions.PermissionDenied(detail=message, code=code)

    def get_project(self):
        """ Returns project instance. """
        return DonationProject.objects.filter(pk=self.kwargs['pk']).first()

    def create_event_log(self, descr, msg):
        """ Creates an event log entry related to the current project. """
        return EventLogEntry.objects.create(project=self.get_project(),
                                            description=descr,
                                            message=msg)

    def delete(self, request, format=None, *args, **kwargs):
        """
        Delete participant related to current project by providing external_id.
        """
        project = self.get_project()
        if not user_is_allowed_to_download(request.user, project):
            self.create_event_log(
                descr='Forbidden Deletion Request',
                msg='Request user is not permitted to delete the project data.'
            )
            msg = 'User does not have access.'
            return Response(status=status.HTTP_403_FORBIDDEN,
                            data={'message': msg})

        external_id = self.kwargs['participant_id']
        try:
            participant = Participant.objects.get(project=project, external_id=external_id)
        except ObjectDoesNotExist:
            msg = f'Participant with id "{external_id}" not found.'
            return Response(status=status.HTTP_404_NOT_FOUND, data={'message': msg})

        participant.delete()

        self.create_event_log(
            descr=f'Participant with external id "{external_id}" successfully deleted.',
            msg='Participant deleted.'
        )

        msg = f'Participant with external id "{external_id}" successfully deleted.'
        return Response(status=status.HTTP_200_OK, data={'message': msg})


class ExceptionAPI(APIView):

    permission_classes = [permissions.AllowAny]

    def post(self, request, format=None, *args, **kwargs):
        """
        Log exception messages received from client.
        """
        project_id = self.kwargs['pk']
        project = DonationProject.objects.get(pk=project_id)

        participant_id = request.session['projects'][f'{project_id}']['participant_id']
        participant = Participant.objects.get(pk=participant_id)

        # Get related blueprint
        blueprint_id = request.data.get('blueprint')
        if blueprint_id:
            blueprint = DonationBlueprint.objects.get(pk=blueprint_id)
        else:
            blueprint = None

        ExceptionLogEntry.objects.create(
            project=project,
            participant=participant,
            exception_type=request.data.get('status_code'),
            message=request.data['message'],
            raised_by=request.data.get('raised_by'),
            blueprint=blueprint
        )

        return Response(None, status=201)



import binascii
import os

from ddm.models.logs import EventLogEntry
from django.db import models
from django.utils import timezone
from rest_framework import exceptions
from rest_framework.authentication import TokenAuthentication, get_authorization_header


class ProjectAccessToken(models.Model):
    """
    Custom authorization token that is linked to a DonationProject.
    Adapted from the Token model as implemented in rest_framework.authtoken.models.Token.
    """
    key = models.CharField(max_length=40, primary_key=True)
    project = models.OneToOneField(
        'DonationProject', related_name='donation_project',
        on_delete=models.CASCADE, verbose_name='Donation Project'
    )
    created = models.DateTimeField(auto_now_add=True)
    expiration_date = models.DateTimeField(blank=True, null=True)

    def save(self, *args, **kwargs):
        if not self.key:
            self.key = self.generate_key()
        return super().save(*args, **kwargs)

    def delete(self, *args, **kwargs):
        EventLogEntry.objects.create(
            project=self.project,
            description='Access Token Deleted'
        )
        super().delete(*args, **kwargs)

    @classmethod
    def generate_key(cls):
        return binascii.hexlify(os.urandom(20)).decode()

    def __str__(self):
        return self.key

    def has_expired(self):
        """ Returns False if token has expired. """
        return self.expiration_date is not None and timezone.now() > self.expiration_date


class ProjectTokenAuthenticator(TokenAuthentication):
    """
    Custom implementation of the rest_framework.authentication.TokenAuthentication
    method that does not require a token to be related to a user. Instead, the
    custom model requires a token to be related to a DonationProject.
    Furthermore, the custom method supports token expiration.

    From original description:
    Clients should authenticate by passing the token key in the "Authorization"
    HTTP header, prepended with the string "Token ".  For example:

        Authorization: Token 401f7ac837da42b97f613d789819ff93537bee6a
    """
    model = ProjectAccessToken

    def authenticate(self, request):
        """
        Adopted from parent model. Added that request is passed to
        authenticate_credentials in order to check if the requested project
        matches the token.
        """
        auth = get_authorization_header(request).split()

        if not auth or auth[0].lower() != self.keyword.lower().encode():
            return None

        if len(auth) == 1:
            msg = 'Invalid token header. No credentials provided.'
            raise exceptions.AuthenticationFailed(msg)
        elif len(auth) > 2:
            msg = 'Invalid token header. Token string should not contain spaces.'
            raise exceptions.AuthenticationFailed(msg)

        try:
            token = auth[1].decode()
        except UnicodeError:
            msg = 'Invalid token header. Token string should not contain invalid characters.'
            raise exceptions.AuthenticationFailed(msg)

        request_url = request.get_full_path()
        project_pk = [x for x in request_url.split('/') if x != ''][0]
        try:
            project_pk = int(project_pk)
        except ValueError:
            msg = 'Invalid project identifier provided.'
            raise exceptions.AuthenticationFailed(msg)

        return self.authenticate_credentials(token, project_pk)

    def authenticate_credentials(self, key, project_pk):
        model = self.get_model()
        try:
            token = model.objects.select_related('project').get(key=key)
        except model.DoesNotExist:
            msg = 'Invalid token.'
            raise exceptions.AuthenticationFailed(msg)

        if token.has_expired():
            msg = 'Token has expired. You can create a new one in the admin backend.'
            raise exceptions.AuthenticationFailed(msg)

        if token.project.pk != project_pk:
            msg = 'The provided token does not belong to the requested project.'
            raise exceptions.AuthenticationFailed(msg)

        return token.project.owner.user, token



import random

from ckeditor.fields import RichTextField

from django.db import models
from django.forms import model_to_dict
from django.template import Context, Template

from polymorphic.models import PolymorphicModel

from ddm.models.core import DataDonation
from ddm.models.logs import ExceptionLogEntry, ExceptionRaisers


class QuestionType(models.TextChoices):
    GENERIC = 'generic', 'Generic Question'
    SINGLE_CHOICE = 'single_choice', 'Single Choice Question'
    MULTI_CHOICE = 'multi_choice', 'Multi Choice Question'
    MATRIX = 'matrix', 'Matrix Question'
    SEMANTIC_DIFF = 'semantic_diff', 'Semantic Differential'
    OPEN = 'open', 'Open Question'
    TRANSITION = 'transition', 'Transition Block'


class QuestionBase(PolymorphicModel):
    project = models.ForeignKey(
        'DonationProject',
        on_delete=models.CASCADE
    )
    blueprint = models.ForeignKey(
        'DonationBlueprint',
        on_delete=models.CASCADE,
        null=True,
        blank=True
    )

    DEFAULT_QUESTION_TYPE = QuestionType.GENERIC
    question_type = models.CharField(
        max_length=20,
        blank=False,
        choices=QuestionType.choices,
        default=DEFAULT_QUESTION_TYPE
    )

    name = models.CharField(max_length=255)
    index = models.PositiveIntegerField(default=1, verbose_name='Page')

    variable_name = models.SlugField(
        max_length=50,
        null=False,
        verbose_name='Variable name for storing response',
        help_text='Will be used in the data export to identify responses to this question.'
    )

    text = RichTextField(null=True, blank=True, config_name='ddm_ckeditor')
    required = models.BooleanField(default=False)

    class Meta:
        ordering = ['index']
        constraints = [
            models.UniqueConstraint(
                fields=['variable_name', 'project'],
                name='unique_varname_per_project'
            ),
        ]

    def __init__(self, *args, **kwargs):
        kwargs['question_type'] = self.DEFAULT_QUESTION_TYPE
        super().__init__(*args, **kwargs)

    def __str__(self):
        return self.name

    def is_general(self):
        return True if self.blueprint is None else False

    def get_config(self, participant_id, view):
        config = self.create_config()
        config = self.render_config_content(config, participant_id, view)
        return config

    def create_config(self):
        config = {
            'question': self.pk,
            'type': self.question_type,
            'index': self.index,
            'text': self.text,
            'required': self.required,
            'items': [],
            'scale': [],
            'options': {}
        }
        return config

    def log_exception(self, msg):
        ExceptionLogEntry.objects.create(
            project=self.project,
            blueprint=self.blueprint,
            raised_by=ExceptionRaisers.SERVER,
            message='Questionnaire Response Processing Exception: ' + msg
        )

    def render_config_content(self, config, participant, view):
        if self.is_general():
            context_data = None
        else:
            data_donation = DataDonation.objects.get(
                participant=participant,
                blueprint=self.blueprint
            )
            context_data = data_donation.get_decrypted_data()
        config['text'] = self.render_text(config['text'], context_data, view)
        for index, item in enumerate(config['items']):
            item['label'] = self.render_text(item['label'], context_data, view)
            item['label_alt'] = self.render_text(item['label_alt'], context_data, view)
        return config

    @staticmethod
    def render_text(text, context, view):
        if text is not None:
            text = '{% load ddm_graphs static %}\n' + text
        template = Template(text)
        return template.render(Context({'data': context, 'view': view}))

    def validate_response(self, response):
        return


class ItemMixin(models.Model):
    randomize_items = models.BooleanField(default=False)

    class Meta:
        abstract = True

    def create_config(self):
        config = super().create_config()
        config = self.add_item_config(config)
        return config

    def add_item_config(self, config):
        items = QuestionItem.objects.filter(question=self)
        for item in items:
            config['items'].append(item.serialize_to_config())

        if self.randomize_items:
            random.shuffle(config['items'])
        return config

    def validate_response(self, response):
        # Validate that all items have a response.
        item_ids = [str(i) for i in list(self.questionitem_set.all().values_list('id', flat=True))]
        if not sorted(item_ids) == sorted(response.keys()):
            if len(item_ids) > len(response.keys()):
                msg = ('Some responses are missing for '
                       f'{self.DEFAULT_QUESTION_TYPE} with ID '
                       f'{self.id}. Got no response for items '
                       f'{[i for i in item_ids if i not in response.keys()]}.')
                self.log_exception(msg)
            elif len(item_ids) < len(response.keys()):
                msg = (f'Got unexpected response keys for '
                       f'{self.DEFAULT_QUESTION_TYPE} with ID {self.id}.'
                       f'Unexpected keys: {[k for k in response.keys() if k not in item_ids]}.')
                self.log_exception(msg)
            else:
                msg = (f'Response does not match the expected items. Items: '
                       f'{sorted(item_ids)}; Keys: {sorted(response.keys())}.')
                self.log_exception(msg)
        return


class ScaleMixin:
    def create_config(self):
        config = super().create_config()
        config = self.add_scale_config(config)
        return config

    def add_scale_config(self, config):
        scale_points = ScalePoint.objects.filter(question=self)
        for point in scale_points:
            config['scale'].append(point.serialize_to_config())
        return config

    def validate_response(self, response):
        super().validate_response(response)
        valid_values = list(self.scalepoint_set.all().values_list('value', flat=True))
        valid_values.append(-99)
        for k, val in response.items():
            if val not in valid_values:
                msg = (f'Got invalid response "{k}: {val}" for multi '
                       f'choice question with ID {self.id}.')
                self.log_exception(msg)
        return


class SingleChoiceQuestion(ItemMixin, QuestionBase):
    DEFAULT_QUESTION_TYPE = QuestionType.SINGLE_CHOICE

    def validate_response(self, response):
        valid_values = list(self.questionitem_set.all().values_list('value', flat=True))
        valid_values.append(-99)
        if response not in valid_values:
            msg = (f'Got invalid response "{response}" for single choice '
                   f'question with ID {self.id}.')
            self.log_exception(msg)
        return


class MultiChoiceQuestion(ItemMixin, QuestionBase):
    DEFAULT_QUESTION_TYPE = QuestionType.MULTI_CHOICE

    def validate_response(self, response):
        super().validate_response(response)
        valid_values = [0, 1, -99]
        for k, val in response.items():
            if val not in valid_values:
                msg = (f'Got invalid response "{k}: {val}" for '
                       f'{self.DEFAULT_QUESTION_TYPE} with ID {self.id}.')
                self.log_exception(msg)
        return


class OpenQuestion(QuestionBase):
    DEFAULT_QUESTION_TYPE = QuestionType.OPEN

    class DisplayOptions(models.TextChoices):
        SMALL = 'small', 'Small'
        LARGE = 'large', 'Large'
    display = models.CharField(
        max_length=20,
        blank=False,
        choices=DisplayOptions.choices,
        default=DisplayOptions.LARGE
    )

    def create_config(self):
        config = super().create_config()
        config['options']['display'] = self.display
        return config

    def validate_response(self, response):
        # TODO: Think about if it's necessary to add some validation steps here.
        return


class MatrixQuestion(ScaleMixin, ItemMixin, QuestionBase):
    DEFAULT_QUESTION_TYPE = QuestionType.MATRIX


class SemanticDifferential(ScaleMixin, ItemMixin, QuestionBase):
    DEFAULT_QUESTION_TYPE = QuestionType.SEMANTIC_DIFF


class Transition(QuestionBase):
    DEFAULT_QUESTION_TYPE = QuestionType.TRANSITION


class QuestionItem(models.Model):
    class Meta:
        ordering = ['index']

    question = models.ForeignKey(
        'QuestionBase',
        on_delete=models.CASCADE
    )
    label = models.CharField(
        max_length=255,
        null=True,
        blank=True
    )
    label_alt = models.CharField(
        max_length=255,
        null=True,
        blank=True,
    )
    index = models.IntegerField()
    value = models.IntegerField()
    randomize = models.BooleanField(default=False)

    def serialize_to_config(self):
        item_config = model_to_dict(self, exclude=['question'])
        return item_config


class ScalePoint(models.Model):
    class Meta:
        ordering = ['index']
        constraints = [
            models.UniqueConstraint(
                fields=['index', 'question'],
                name='unique_index_per_question'
            ),
        ]

    question = models.ForeignKey(
        'QuestionBase',
        on_delete=models.CASCADE
    )
    index = models.IntegerField()
    label = models.CharField(
        max_length=100,
        null=True,
        blank=True
    )
    value = models.IntegerField()
    add_border = models.BooleanField(default=False)

    def serialize_to_config(self):
        scale_config = model_to_dict(self, exclude=['question', 'add_border'])
        return scale_config



from django.views.decorators.debug import sensitive_variables
from rest_framework import serializers
from rest_framework.fields import empty

from ddm.models.core import DataDonation, DonationProject, QuestionnaireResponse


class ProjectSerializer(serializers.HyperlinkedModelSerializer):

    class Meta:
        model = DonationProject
        fields = ['pk', 'name', 'date_created']


class SerializerDecryptionMixin:
    """
    Allows to pass the secret for the decryption of super secret projects to
    the serializer on init.
    """
    @sensitive_variables()
    def __init__(self, instance=None, data=empty, **kwargs):
        self.secret = kwargs.pop('secret', None)
        super().__init__(instance=instance, data=data, **kwargs)

    @sensitive_variables()
    def get_data(self, obj):
        kwargs = {}
        if self.secret is not None:
            kwargs['secret'] = self.secret
        return obj.get_decrypted_data(**kwargs)


class DonationSerializer(SerializerDecryptionMixin, serializers.HyperlinkedModelSerializer):
    project = serializers.IntegerField(source='project.id')
    data = serializers.SerializerMethodField()
    participant = serializers.IntegerField(source='participant.id')

    class Meta:
        model = DataDonation
        fields = ['time_submitted', 'consent', 'status', 'data', 'project', 'participant']


class ResponseSerializer(SerializerDecryptionMixin, serializers.HyperlinkedModelSerializer):
    project = serializers.IntegerField(source='project.id')
    data = serializers.CharField(source='get_decrypted_data')
    participant = serializers.IntegerField(source='participant.id')

    class Meta:
        model = QuestionnaireResponse
        fields = ['time_submitted', 'data', 'project', 'participant']



XXXXXXXXX XXXXX
XXXXX XXXXXXXX
  XXXXXX
    XXXXX XXXXXXXXXXXXXXXX
    XXXXX XXXXXXXXXXXXXXXXXXXXXXXXXXXX XXXXXXXXXXXXXXXXXX
    XXXXX XXXXXXXXXXXXXXX XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
    XXXX XXXXX XXXXXXXXXX XXXXXXXXX XXXXXXXX XXXXXXXXXXXXXXX XXX
    XXXXXXXXXX XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX XXXXXXXXXX
  XXXXXXX
  XXXXXX
    XXXXXXXXXX
      XXXXXXXXXXXXX XXXXX XXX XXX XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX XX XXXXXXX XXXX XXXXXXXX XXXXXXX XXXXXXXXXX XXXXXXXX XXXXXX XXXXXX XX XX XXXXXXXXXXXXXXXXXX
    XXXXXXXXXXX
    XXXX XXXXXXXXXXXXXXX
    XXXX XXXXX XXXXX XXXX XX XXXX XXXXXXXX XXX
  XXXXXXX
XXXXXXX



from django.db import models
from django.utils import timezone


class ExceptionRaisers(models.TextChoices):
    SERVER = 'server'
    CLIENT = 'client'


class ExceptionLogEntry(models.Model):
    date = models.DateTimeField(default=timezone.now)
    project = models.ForeignKey(
        'DonationProject',
        on_delete=models.CASCADE
    )
    participant = models.ForeignKey(
        'Participant',
        null=True,
        on_delete=models.SET_NULL
    )
    blueprint = models.ForeignKey(
        'DonationBlueprint',
        null=True,
        on_delete=models.CASCADE
    )
    raised_by = models.CharField(
        max_length=20,
        choices=ExceptionRaisers.choices,
        blank=True
    )
    exception_type = models.IntegerField(
        null=True
    )
    message = models.TextField()


class EventLogEntry(models.Model):
    date = models.DateTimeField(default=timezone.now)
    project = models.ForeignKey(
        'DonationProject',
        on_delete=models.CASCADE
    )
    description = models.CharField(max_length=100)
    message = models.TextField()



from django.apps import apps
from django.core import checks


def check_dependencies(**kwargs):
    """
    Check that the dependencies of ddm.stats are correctly installed.
    """
    if not apps.is_installed('ddm.stats'):
        return []
    errors = []
    app_dependencies = ['ddm']
    for app_name in app_dependencies:
        if not apps.is_installed(app_name):
            errors.append(
                checks.Error(
                    f'{app_name} must be in INSTALLED_APPS in order to use the '
                    f'ddm stats application.'
                )
            )
    return errors



from django.apps import AppConfig
from django.core import checks

from ddm.stats.checks import check_dependencies


class DdmStatsConfig(AppConfig):
    name = 'ddm.stats'
    label = 'ddmstats'
    verbose_name = 'DDM Stats'

    def ready(self):
        checks.register(check_dependencies, checks.Tags.admin)



# Generated by Django 3.2.13 on 2023-01-04 12:25

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ddm', '0022_auto_20230103_1433'),
    ]

    operations = [
        migrations.RenameField(
            model_name='donationproject',
            old_name='intro_text',
            new_name='briefing_text',
        ),
        migrations.RenameField(
            model_name='donationproject',
            old_name='outro_text',
            new_name='debriefing_text',
        ),
        migrations.RemoveField(
            model_name='donationproject',
            name='collaborators',
        ),
        migrations.AddField(
            model_name='donationproject',
            name='briefing_consent_enabled',
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name='donationproject',
            name='briefing_consent_label_no',
            field=models.CharField(blank=True, max_length=255),
        ),
        migrations.AddField(
            model_name='donationproject',
            name='briefing_consent_label_yes',
            field=models.CharField(blank=True, max_length=255),
        ),
        migrations.AlterField(
            model_name='donationproject',
            name='expected_url_parameters',
            field=models.CharField(blank=True, default='', help_text='Separate multiple parameters with a semikolon (";"). Semikolons are not allowed as part of the expected url parameters.', max_length=500, verbose_name='Expected URL parameter'),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name='donationproject',
            name='redirect_target',
            field=models.CharField(blank=True, default='', help_text='Always include <i>http://</i> or <i>https://</i> in the redirect target. If URL parameter extraction is enabled for this project, you can include the extracted URL parameters in the redirect target as follows: "https://redirect.me/?redirectpara=<b>{{URLParameter}}</b>".', max_length=2000, verbose_name='Redirect target'),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name='fileuploader',
            name='upload_type',
            field=models.CharField(choices=[('zip file', 'Zip File'), ('single file', 'Single File')], default='single file', max_length=20, verbose_name='Upload type'),
        ),
    ]



from django.contrib.auth import get_user_model
from django.test import Client, TestCase, override_settings
from django.urls import reverse
from django.utils import timezone

from ddm.models.core import (
    DataDonation, Participant, ResearchProfile, DonationProject,
    DonationBlueprint, FileUploader
)
from ddm.models.questions import OpenQuestion


User = get_user_model()


@override_settings(DDM_SETTINGS={'EMAIL_PERMISSION_CHECK':  r'.*(\.|@)mail\.com$', })
class ParticipationFlowBaseTestCase(TestCase):

    @classmethod
    def setUpTestData(cls):
        user = User.objects.create_user(**{
            'username': 'owner', 'password': '123', 'email': 'owner@mail.com'
        })
        profile = ResearchProfile.objects.create(user=user)

        cls.project_base = DonationProject.objects.create(
            name='Base Project', slug='base', owner=profile)

        cls.project_alt = DonationProject.objects.create(
            name='Alt Project', slug='alt', owner=profile)

        file_uploader = FileUploader.objects.create(
            project=cls.project_base,
            name='basic file uploader',
            upload_type=FileUploader.UploadTypes.SINGLE_FILE
        )

        cls.blueprint = DonationBlueprint.objects.create(
            project=cls.project_base,
            name='donation blueprint',
            expected_fields='"a", "b"',
            file_uploader=file_uploader
        )

        OpenQuestion.objects.create(
            project=cls.project_base,
            blueprint=cls.blueprint,
            name='open question',
            variable_name='open_question'
        )

        # URLs for project with questionnaire.
        slug_base = cls.project_base.slug
        cls.briefing_url = reverse('briefing', args=[slug_base])
        cls.dd_url = reverse('data-donation', args=[slug_base])
        cls.quest_url = reverse('questionnaire', args=[slug_base])
        cls.debriefing_url = reverse('debriefing', args=[slug_base])

        # URLs for project without questionnaire.
        slug_alt = cls.project_alt.slug
        cls.briefing_url_no_quest = reverse('briefing', args=[slug_alt])
        cls.dd_url_no_quest = reverse('data-donation', args=[slug_alt])
        cls.quest_url_no_quest = reverse('questionnaire', args=[slug_alt])
        cls.debriefing_url_no_quest = reverse('debriefing', args=[slug_alt])

        # URLs for non-existing project.
        cls.briefing_url_invalid = reverse('briefing', args=['nope'])
        cls.dd_url_invalid = reverse('data-donation', args=['nope'])
        cls.quest_url_invalid = reverse('questionnaire', args=['nope'])
        cls.debriefing_url_invalid = reverse('debriefing', args=['nope'])

    def initialize_project_and_session(self):
        self.client = Client()
        self.client.get(self.briefing_url)
        self.client.get(self.briefing_url_no_quest)

    def create_data_donation(self):
        # Create a data donation for project 1.
        session = self.client.session
        participant_id = session[f'project-{self.project_base.pk}']['participant_id']
        participant = Participant.objects.get(pk=int(participant_id))
        DataDonation.objects.create(
            project=self.project_base,
            blueprint=self.blueprint,
            participant=participant,
            time_submitted=timezone.now(),
            consent=True,
            status='{}',
            data='{}'
        )


class TestBriefingView(ParticipationFlowBaseTestCase):

    def setUp(self):
        super().setUp()

    def test_project_briefing_view_registers_project(self):
        self.client.get(self.briefing_url)
        session = self.client.session
        expected_keys = ['last_completed_step', 'participant_id']
        actual_keys = set(session[f'project-{self.project_base.pk}'].keys())
        assert set(expected_keys).issubset(actual_keys)

    def test_project_briefing_view_registers_participant(self):
        nr_participants_before = len(Participant.objects.all())
        self.client.get(self.briefing_url)
        self.assertGreater(len(Participant.objects.all()), nr_participants_before)

        project_session = self.client.session[f'project-{self.project_base.pk}']
        self.assertIsNotNone(project_session['participant_id'])

    def test_project_briefing_view_GET_valid_url(self):
        response = self.client.get(self.briefing_url)
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, 'ddm/public/briefing.html')

    def test_project_briefing_view_GET_invalid_url(self):
        response = self.client.get(self.briefing_url_invalid, follow=True)
        self.assertEqual(response.status_code, 404)

    def test_project_briefing_POST(self):
        response = self.client.post(self.briefing_url)
        project_session = self.client.session[f'project-{self.project_base.pk}']
        self.assertEqual(project_session['last_completed_step'], 0)
        self.assertEqual(response.status_code, 302)
        self.assertRedirects(response, self.dd_url)

    def test_project_briefing_consent_yes_POST(self):
        self.project_base.briefing_consent_enabled = True
        self.project_base.save()
        self.client.get(self.briefing_url)
        response = self.client.post(
            self.briefing_url, {'briefing_consent': '1'}, follow=True)
        project_session = self.client.session[f'project-{self.project_base.pk}']

        self.assertEqual(project_session['last_completed_step'], 0)
        self.assertRedirects(response, self.dd_url)

        # Assert consent is correctly stored in participant data.
        participant_id = project_session['participant_id']
        participant = Participant.objects.get(pk=participant_id)
        self.assertEqual(participant.extra_data['briefing_consent'], '1')

    def test_project_briefing_consent_no_POST(self):
        self.project_base.briefing_consent_enabled = True
        self.project_base.save()
        self.client.get(self.briefing_url)
        response = self.client.post(self.briefing_url, {'briefing_consent': '0'})
        project_session = self.client.session[f'project-{self.project_base.pk}']

        self.assertEqual(project_session['last_completed_step'], 2)
        self.assertRedirects(response, self.debriefing_url)

        # Assert consent is correctly stored in participant data.
        participant_id = project_session['participant_id']
        participant = Participant.objects.get(pk=participant_id)
        self.assertEqual(participant.extra_data['briefing_consent'], '0')

    def test_project_briefing_consent_invalid_POST(self):
        self.project_base.briefing_consent_enabled = True
        self.project_base.save()
        self.client.get(self.briefing_url)
        response = self.client.post(self.briefing_url)
        project_session = self.client.session[f'project-{self.project_base.pk}']

        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, 'ddm/public/briefing.html')


class TestDonationView(ParticipationFlowBaseTestCase):

    def setUp(self):
        super().setUp()
        self.initialize_project_and_session()
        self.create_data_donation()
        session = self.client.session
        session[f'project-{self.project_base.pk}']['last_completed_step'] = 0
        session.save()

    def test_data_donation_GET_valid_url(self):
        response = self.client.get(self.dd_url)
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, 'ddm/public/data_donation.html')

    def test_data_donation_GET_invalid_url(self):
        response = self.client.get(self.dd_url_invalid, follow=True)
        self.assertEqual(response.status_code, 404)

    def test_data_donation_POST_redirect(self):
        response = self.client.post(self.dd_url)
        self.assertEqual(response.status_code, 302)
        self.assertRedirects(response, self.quest_url)


class TestQuestionnaireView(ParticipationFlowBaseTestCase):

    def setUp(self):
        super().setUp()
        self.initialize_project_and_session()
        self.create_data_donation()
        session = self.client.session
        session[f'project-{self.project_base.pk}']['last_completed_step'] = 1
        session[f'project-{self.project_alt.pk}']['last_completed_step'] = 1
        session.save()

    def test_questionnaire_GET_valid_url(self):
        response = self.client.get(self.quest_url)
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, 'ddm/public/questionnaire.html')

    def test_questionnaire_GET_invalid_url(self):
        response = self.client.get(self.quest_url_invalid, follow=True)
        self.assertEqual(response.status_code, 404)

    def test_questionnaire_GET_no_questionnaire(self):
        response = self.client.get(self.quest_url_no_quest)
        self.assertRedirects(response, self.debriefing_url_no_quest)

    def test_questionnaire_POST_redirect(self):
        response = self.client.post(self.quest_url)
        self.assertEqual(response.status_code, 302)
        self.assertRedirects(response, self.debriefing_url)


class TestDebriefingView(ParticipationFlowBaseTestCase):

    def setUp(self):
        super().setUp()
        self.initialize_project_and_session()
        session = self.client.session
        session[f'project-{self.project_base.pk}']['last_completed_step'] = 2
        session.save()

    def test_project_debriefing_GET_valid_url(self):
        response = self.client.get(self.debriefing_url)
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, 'ddm/public/debriefing.html')

    def test_project_debriefing_GET_invalid_url(self):
        response = self.client.get(self.debriefing_url_invalid, follow=True)
        self.assertEqual(response.status_code, 404)


# TODO: Add Class TestViewsRerouting(TestCase)
class TestRedirect(ParticipationFlowBaseTestCase):

    def setUp(self):
        super().setUp()
        self.initialize_project_and_session()
        self.project_base.redirect_enabled = True
        self.project_base.url_parameter_enabled = True
        self.project_base.save()

    def test_redirect_single_parameter(self):
        self.project_base.expected_url_parameters = 'testparam'
        self.project_base.redirect_target = 'http://test.test/?para={{testparam}}'
        self.project_base.save()

        self.client.get(self.briefing_url + '?testparam=test')

        session = self.client.session
        session[f'project-{self.project_base.pk}']['last_completed_step'] = 2
        session.save()

        response = self.client.get(self.debriefing_url)
        self.assertEqual(
            response.context['redirect_target'],
            'http://test.test/?para=test'
        )

    def test_redirect_multiple_parameters(self):
        self.project_base.expected_url_parameters = 'testparam;testparam2'
        self.project_base.redirect_target = 'http://test.test/?para={{testparam}}&para2={{testparam2}}'
        self.project_base.save()
        self.client.get(self.briefing_url + '?testparam=test&testparam2=test2')

        session = self.client.session
        session[f'project-{self.project_base.pk}']['last_completed_step'] = 2
        session.save()

        response = self.client.get(self.debriefing_url)
        self.assertEqual(
            response.context['redirect_target'],
            'http://test.test/?para=test&para2=test2'
        )



# Generated by Django 3.2.13 on 2022-12-16 13:28

from django.db import migrations
import django.db.models.manager


class Migration(migrations.Migration):

    dependencies = [
        ('ddm', '0019_auto_20221212_2121'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='questionbase',
            options={'base_manager_name': 'non_polymorphic', 'ordering': ['index']},
        ),
        migrations.AlterModelManagers(
            name='matrixquestion',
            managers=[
                ('non_polymorphic', django.db.models.manager.Manager()),
            ],
        ),
        migrations.AlterModelManagers(
            name='multichoicequestion',
            managers=[
                ('non_polymorphic', django.db.models.manager.Manager()),
            ],
        ),
        migrations.AlterModelManagers(
            name='openquestion',
            managers=[
                ('non_polymorphic', django.db.models.manager.Manager()),
                ('objects', django.db.models.manager.Manager()),
            ],
        ),
        migrations.AlterModelManagers(
            name='questionbase',
            managers=[
                ('non_polymorphic', django.db.models.manager.Manager()),
            ],
        ),
        migrations.AlterModelManagers(
            name='semanticdifferential',
            managers=[
                ('non_polymorphic', django.db.models.manager.Manager()),
            ],
        ),
        migrations.AlterModelManagers(
            name='singlechoicequestion',
            managers=[
                ('non_polymorphic', django.db.models.manager.Manager()),
            ],
        ),
        migrations.AlterModelManagers(
            name='transition',
            managers=[
                ('non_polymorphic', django.db.models.manager.Manager()),
                ('objects', django.db.models.manager.Manager()),
            ],
        ),
    ]



from django.apps import AppConfig


class DdmConfig(AppConfig):
    default_auto_field = 'django.db.models.AutoField'
    name = 'ddm'
    verbose_name = 'Data Donation Module'

    def ready(self):
        from ddm import signals



import datetime
import json
import os
import random
import string

from ckeditor.fields import RichTextField

from django.conf import settings
from django.core.exceptions import ValidationError, ObjectDoesNotExist
from django.core.validators import RegexValidator, MinValueValidator, MinLengthValidator
from django.db import models
from django.db.models import Avg, F, ImageField
from django.urls import reverse
from django.utils import timezone
from django.utils.safestring import mark_safe
from django.views.decorators.debug import sensitive_variables

from ddm.models.auth import ProjectAccessToken
from ddm.models.encryption import Encryption, ModelWithEncryptedData
from ddm.models.logs import ExceptionLogEntry, ExceptionRaisers, EventLogEntry


class ResearchProfile(models.Model):
    user = models.OneToOneField(
        settings.AUTH_USER_MODEL,
        on_delete=models.CASCADE,
        related_name='ddm_research_profile',
    )
    active = models.BooleanField(default=True)
    created = models.DateTimeField('date registered', default=timezone.now)
    ignore_email_restriction = models.BooleanField(default=False)


def project_header_dir_path(instance, filename):
    return f'project_{instance.pk}/headers/{filename}'


class DonationProject(models.Model):
    # Basic information for internal organization.
    name = models.CharField(max_length=50)
    date_created = models.DateTimeField(default=timezone.now)
    owner = models.ForeignKey(
        'ResearchProfile',
        related_name='project_owner',
        on_delete=models.SET_NULL,
        null=True
    )

    # Public information.
    contact_information = RichTextField(
        null=True,
        blank=False,
        verbose_name='Contact information',
        help_text=(
            'Please provide the contact information of the person responsible '
            'for the conduction of this study. The contact information will be '
            'accessible for participants during the data donation.'
        ),
        config_name='ddm_ckeditor'
    )
    data_protection_statement = RichTextField(
        null=True,
        blank=False,
        verbose_name='Data Protection Statement',
        help_text=(
            'Please provide a data protection statement for your data donation '
            'collection. This should include the purpose for which the data is '
            'collected, how it will be stored, and who will have access to the data.'
        ),
        config_name='ddm_ckeditor'
    )

    # Information affecting participation flow.
    slug = models.SlugField(unique=True, verbose_name='External Project Slug')
    briefing_text = RichTextField(
        null=True, blank=True,
        verbose_name='Welcome Page Text',
        config_name='ddm_ckeditor'
    )
    briefing_consent_enabled = models.BooleanField(default=False)
    briefing_consent_label_yes = models.CharField(max_length=255, blank=True)
    briefing_consent_label_no = models.CharField(max_length=255, blank=True)
    debriefing_text = RichTextField(
        null=True, blank=True,
        verbose_name='End Page Text',
        config_name='ddm_ckeditor'
    )

    # Appearance settings.
    img_header_left = ImageField(
        upload_to=project_header_dir_path,
        null=True,
        blank=True,
        verbose_name='Header Image Left'
    )
    img_header_right = ImageField(
        upload_to=project_header_dir_path,
        null=True,
        blank=True,
        verbose_name='Header Image Right'
    )

    # Access settings.
    public_key = models.BinaryField()
    super_secret = models.BooleanField(default=False)

    # Redirect settings.
    redirect_enabled = models.BooleanField(default=False, verbose_name='Redirect enabled')
    redirect_target = models.CharField(
        max_length=2000,
        blank=True,
        verbose_name='Redirect target',
        help_text=mark_safe(
            'Always include <i>http://</i> or <i>https://</i> in the redirect target. '
            'If URL parameter extraction is enabled for this project, you can '
            'include the extracted URL parameters in the redirect target as follows: '
            '"https://redirect.me/?redirectpara=<b>{{URLParameter}}</b>".')
    )

    # URL parameter extraction settings.
    url_parameter_enabled = models.BooleanField(default=False, verbose_name='URL parameter extraction enabled')
    expected_url_parameters = models.CharField(
        max_length=500,
        blank=True,
        verbose_name='Expected URL parameter',
        help_text='Separate multiple parameters with a semikolon (";"). '
                  'Semikolons are not allowed as part of the expected url parameters.'
    )

    @sensitive_variables()
    def __init__(self, *args, **kwargs):
        self.secret_key = settings.SECRET_KEY
        if 'secret_key' in kwargs:
            self.secret_key = kwargs['secret_key']
            kwargs.pop('secret_key')
        super().__init__(*args, **kwargs)

    def __str__(self):
        return self.name

    def get_absolute_url(self):
        return reverse('project-detail', args=[str(self.id)])

    def clean_file_on_reupload(self):
        """
        Deletes previous file on model FileFields or ImageFields if a file
        is re-uploaded.
        """
        i = DonationProject.objects.get(pk=self.pk)
        for field in i._meta.fields:
            if field.get_internal_type() in ['FileField', 'ImageField']:
                if getattr(i, field.name) != getattr(self, field.name):
                    file_attr = getattr(i, field.name)
                    try:
                        if os.path.isfile(file_attr.path):
                            os.remove(file_attr.path)
                    except ValueError:
                        pass
        return

    @sensitive_variables()
    def save(self, *args, **kwargs):
        if not self.pk:
            if self.owner is None:
                raise ValidationError(
                    'DonationProject.owner cannot be null on create.')
        else:
            self.clean_file_on_reupload()

        if not self.public_key:
            self.public_key = Encryption.get_public_key(self.secret, str(self.date_created))
        super().save(*args, **kwargs)

    @property
    def secret(self):
        return self.secret_key

    @secret.setter
    def secret(self, value):
        self.secret_key = value

    def delete(self, using=None, keep_parents=False):
        # Manually delete related questions to circumvent foreign key integrity
        # problem caused by QuestionBase models inheriting from django-polymorphic.
        # (see https://github.com/django-polymorphic/django-polymorphic/issues/34#issuecomment-1027866872)
        for question in self.questionbase_set.all():
            question.delete()
        return super().delete(using, keep_parents)

    def get_statistics(self):
        participants = Participant.objects.filter(project=self)
        statistics = {
            'n_started': participants.count(),
            'n_completed': participants.filter(completed=True).count(),
            'completion_rate': participants.filter(completed=True).count() / participants.count() if participants.count() > 0 else 0,
            'n_donations': DataDonation.objects.filter(project=self).count(),
            'n_errors': ExceptionLogEntry.objects.filter(project=self).count(),
            'average_time': str(participants.filter(completed=True).aggregate(v=Avg(F('end_time')-F('start_time')))['v']).split(".")[0]
        }
        return statistics

    def get_questionnaire_config(self, participant, view):
        """
        Returns a dictionary containing all information to render the
        questionnaire for a given participant.
        """
        q_config = {}
        questions = self.questionbase_set.all()
        for question in questions:
            if question.is_general():
                q_config[question.pk] = question.get_config(participant, view)
            else:
                try:
                    donation = DataDonation.objects.get(
                        blueprint=question.blueprint,
                        participant=participant
                    )
                except ObjectDoesNotExist:
                    msg = ('Questionnaire Rendering Exception: No donation '
                           f'found for participant {participant.pk} and '
                           f'blueprint {question.blueprint.pk}.')
                    ExceptionLogEntry.objects.create(
                        project=self,
                        raised_by=ExceptionRaisers.SERVER,
                        message=msg
                    )
                    continue

                if donation.consent:
                    q_config[question.pk] = question.get_config(participant, view)
        return q_config

    def get_expected_url_parameters(self):
        return self.expected_url_parameters.split(';')

    def get_token(self):
        return ProjectAccessToken.objects.filter(project=self).first()

    def create_token(self, expiration_days=None):
        token = self.get_token()
        if token:
            token.delete()
        if expiration_days:
            expiration_date = timezone.now() + datetime.timedelta(days=expiration_days)
        else:
            expiration_date = None
        EventLogEntry.objects.create(
            project=self, description='New Access Token Created')
        return ProjectAccessToken.objects.create(
            project=self, expiration_date=expiration_date)


def get_extra_data_default():
    """ Return default value for Participant.extra_data. """
    return dict(url_param=dict())


def create_asciidigits_id():
    return ''.join(random.choices(string.ascii_letters + string.digits, k=24))


class Participant(models.Model):
    project = models.ForeignKey('DonationProject', on_delete=models.CASCADE)

    external_id = models.CharField(
        unique=True, null=False,
        max_length=24,
        validators=[MinLengthValidator(24)]
    )

    # Participation statistics.
    start_time = models.DateTimeField()
    end_time = models.DateTimeField(null=True)
    completed = models.BooleanField(default=False)

    extra_data = models.JSONField(default=get_extra_data_default)

    def save(self, *args, **kwargs):
        if self.pk is None:
            new_external_id = create_asciidigits_id()
            while len(Participant.objects.filter(external_id=new_external_id)) != 0:
                new_external_id = create_asciidigits_id()
            self.external_id = new_external_id
        super().save(*args, **kwargs)


class QuestionnaireResponse(ModelWithEncryptedData):
    # Will only ever be deleted, when the project is deleted.
    project = models.ForeignKey('DonationProject', on_delete=models.CASCADE)
    participant = models.ForeignKey('Participant', on_delete=models.CASCADE)
    time_submitted = models.DateTimeField(default=timezone.now)
    data = models.BinaryField()


COMMA_SEPARATED_STRINGS_VALIDATOR = RegexValidator(
    r'^((["][^"]+["]))(\s*,\s*((["][^"]+["])))*[,\s]*$',
    message=(
        'Field must contain one or multiple comma separated strings. '
        'Strings must be enclosed in double quotes ("string").'
    )
)


class FileUploader(models.Model):
    name = models.CharField(max_length=250)
    project = models.ForeignKey('DonationProject', on_delete=models.CASCADE)
    index = models.PositiveIntegerField()

    class UploadTypes(models.TextChoices):
        ZIP_FILE = 'zip file'
        SINGLE_FILE = 'single file'

    upload_type = models.CharField(
        max_length=20,
        choices=UploadTypes.choices,
        default=UploadTypes.SINGLE_FILE,
        verbose_name='Upload type',
    )

    def __str__(self):
        return self.name

    def delete(self, *args, **kwargs):
        """ This model has a post_delete signal processor (see signals.py). """
        super().delete(*args, **kwargs)

    def get_configs(self):
        configs = {
            'upload_type': self.upload_type,
            'name': self.name,
            'blueprints': [bp.get_config() for bp in self.donationblueprint_set.all()],
            'instructions': [{'index': i.index, 'text': i.text} for i in self.donationinstruction_set.all()]
        }
        return configs

    def save(self, *args, **kwargs):
        if self.index is None:
            uploader_indices = self.project.fileuploader_set.all().values_list('index', flat=True)
            self.index = 1 if not uploader_indices else max(uploader_indices) + 1
        super().save(*args, **kwargs)


class DonationBlueprint(models.Model):
    project = models.ForeignKey(
        'DonationProject',
        on_delete=models.CASCADE
    )
    name = models.CharField(max_length=250)

    class FileFormats(models.TextChoices):
        JSON_FORMAT = 'json'
        CSV_FORMAT = 'csv'
        # HTML_FORMAT = 'html',
        # XLSX_FORMAT = 'xlsx',

    exp_file_format = models.CharField(
        max_length=10,
        choices=FileFormats.choices,
        default=FileFormats.JSON_FORMAT,
        verbose_name='Expected file format',
    )

    csv_delimiter = models.CharField(
        max_length=10,
        default="",
        blank=True
    )

    expected_fields = models.TextField(
        null=False,
        blank=False,
        validators=[COMMA_SEPARATED_STRINGS_VALIDATOR],
        help_text='Put the field names in double quotes (") and separate them with commas ("Field A", "Field B").'
    )

    file_uploader = models.ForeignKey(
        'FileUploader',
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        verbose_name='Associated File Uploader',
    )
    regex_path = models.TextField(null=True, blank=True)

    def __str__(self):
        return self.name

    def get_absolute_url(self):
        return reverse('blueprint-edit', args=[str(self.project_id), str(self.id)])

    def get_slug(self):
        return 'blueprint'

    def get_config(self):
        # TODO: Change 'f_expected' and 'f_extract' to something more meaningful
        config = {
            'id': self.pk,
            'name': self.name,
            'format': self.exp_file_format,
            'f_expected': json.loads("[" + str(self.expected_fields) + "]"),
            'f_extract': self.get_fields_to_extract(),
            'regex_path': self.regex_path,
            'filter_rules': self.get_filter_rules(),
            'csv_delimiter': self.csv_delimiter
        }
        return config

    def get_instructions(self):
        return [{'index': i.index, 'text': i.text} for i in self.donationinstruction_set.all()]

    def get_associated_questions(self):
        return self.questionbase_set.all()

    def get_filter_rules(self):
        return [r.get_rule_config() for r in self.processingrule_set.all().order_by('execution_order')]

    def get_fields_to_extract(self):
        fields = set()
        for r in self.processingrule_set.all():
            fields.add(r.field)
        return list(fields)

    def process_donation(self, data, participant):
        if self.validate_donation(data):
            self.create_donation(data, participant)
        else:
            msg = ('Data Donation Processing Exception: Donation validation '
                   f'failed for blueprint {self.pk}')
            ExceptionLogEntry.objects.create(
                project=self.project,
                blueprint=self,
                raised_by=ExceptionRaisers.SERVER,
                message=msg
            )
        return

    def validate_donation(self, data):
        # Check if all expected fields are in response.
        response_fields = ['consent', 'extracted_data', 'status']
        if not all(k in data for k in response_fields):
            msg = ('Data Donation Processing Exception: Donation data for '
                   f'Donation Blueprint {self.pk} does not contain the '
                   f'expected information. Expected fields: {response_fields}; '
                   f'Present fields: {data.keys()}.')
            ExceptionLogEntry.objects.create(
                project=self.oproject,
                blueprint=self,
                raised_by=ExceptionRaisers.SERVER,
                message=msg
            )
            return False

        # TODO: Add other validation steps? - e.g., validation of extracted fields
        return True

    def create_donation(self, data, participant):
        DataDonation.objects.create(
            project=self.project,
            blueprint=self,
            participant=participant,
            consent=data['consent'],
            status=data['status'],
            data=data['extracted_data']
        )
        return


class ProcessingRule(models.Model):
    """
    A processing rule that is executed on the specified blueprint field,
    during the file upload in vue.
    Filters or matches values.
    TODO: Improve this description.
    """
    blueprint = models.ForeignKey(
        'DonationBlueprint',
        null=False,
        blank=False,
        on_delete=models.CASCADE
    )

    name = models.CharField(max_length=250)

    field = models.TextField(null=False, blank=False)
    execution_order = models.IntegerField()

    class ComparisonOperators(models.TextChoices):
        EQUAL = '==', 'Equal (==)'
        NOT_EQUAL = '!=', 'Not Equal (!=)'
        GREATER = '>', 'Greater than (>)'
        SMALLER = '<', 'Smaller than (<)'
        GREATER_OR_EQUAL = '>=', 'Greater than or equal (>=)'
        SMALLER_OR_EQUAL = '<=', 'Smaller than or equal (<=)'
        REGEX = 'regex', 'Regex (removes matches)'

    comparison_operator = models.CharField(
        max_length=10,
        blank=True,
        null=True,
        choices=ComparisonOperators.choices,
        default=None
    )
    comparison_value = models.TextField(blank=True)

    def get_rule_config(self):
        """
        Return a configuration dict for the processing rule:
        {
            'field': 'field_name',
            'comparison_operator': '==' | '!=' | '>' | '<' | '>=' | '<=' | 'regex' | None,
            'comparison_value': '123' | Regex-String | None
        }
        """
        return {
            'field': self.field,
            'comparison_operator': self.comparison_operator,
            'comparison_value': self.comparison_value
        }


class DataDonation(ModelWithEncryptedData):
    project = models.ForeignKey(
        'DonationProject',
        on_delete=models.CASCADE
    )
    blueprint = models.ForeignKey(
        'DonationBlueprint',
        null=True,
        on_delete=models.SET_NULL
    )
    participant = models.ForeignKey(
        'Participant',
        on_delete=models.CASCADE
    )
    time_submitted = models.DateTimeField(default=timezone.now)
    consent = models.BooleanField(default=False)
    status = models.JSONField()  # 'success' if data was successfully donated; 'pending' otherwise
    data = models.BinaryField()


# TODO: Outsource to separate model.py file when instruction database is added.
class DonationInstruction(models.Model):
    text = RichTextField(null=True, blank=True, config_name='ddm_ckeditor')
    index = models.PositiveIntegerField(default=1, validators=[MinValueValidator(1)])

    file_uploader = models.ForeignKey(
        'FileUploader',
        null=True,
        blank=False,
        on_delete=models.CASCADE,
        verbose_name='Associated File Uploader',
    )

    class Meta:
        ordering = ['index']
        constraints = [
            models.UniqueConstraint(
                fields=['index', 'file_uploader'],
                name='unique_index_per_file_uploader'
            ),
        ]

    def clean(self):
        # Ensure that index of instruction page is not greater than set of existing instructions + 1.
        n_instructions = self.file_uploader.donationinstruction_set.all().count()
        if self.pk:
            if self.index > n_instructions:
                raise ValidationError(
                    f'Index must be in range 1 to {n_instructions}.'
                )
        else:
            if self.index > (n_instructions + 1):
                raise ValidationError(
                    f'Index must be in range 1 to {n_instructions + 1}.'
                )
        super().clean()

    def save(self, *args, **kwargs):
        if kwargs.pop('ignore_index_check', False):
            return super().save()

        # TODO: Optimize and prettify the following part:
        initial_index = DonationInstruction.objects.get(pk=self.pk).index if self.pk else None
        index_taken = self.file_uploader.donationinstruction_set.filter(index=self.index).exclude(pk=self.pk).exists()
        if index_taken and (self.index != initial_index):

            # Account for unique constraint by doing a "proxy"-save to free index.
            target_index = self.index
            self.index = self.file_uploader.donationinstruction_set.count() + 5
            super().save()

            # Change indices of involved objects:
            queryset = self.file_uploader.donationinstruction_set.exclude(pk=self.pk)
            if initial_index is None:
                queryset = queryset.filter(index__gte=target_index).order_by('-index')
                for q in queryset:
                    q.index += 1
                for q in queryset:
                    q.save(ignore_index_check=True)
            elif target_index < initial_index:
                queryset = queryset.filter(index__gte=target_index, index__lt=initial_index).order_by('-index')
                for q in queryset:
                    q.index += 1
                for q in queryset:
                    q.save(ignore_index_check=True)
            elif target_index > initial_index:
                queryset = queryset.filter(index__gt=initial_index, index__lte=target_index).order_by('index')
                for q in queryset:
                    q.index -= 1
                for q in queryset:
                    q.save(ignore_index_check=True)

            # Revert "proxy"-save.
            self.index = target_index
            return super().save()

        return super().save()

    def delete(self, *args, **kwargs):
        """ This model has a post_delete signal processor (see signals.py). """
        super().delete(*args, **kwargs)



# Generated by Django 3.2.13 on 2022-12-22 08:25

import ckeditor.fields
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ddm', '0020_auto_20221216_1428'),
    ]

    operations = [
        migrations.AddField(
            model_name='blueprintcontainer',
            name='index',
            field=models.PositiveIntegerField(default=1, null=True),
        ),
        migrations.AddField(
            model_name='donationblueprint',
            name='index',
            field=models.PositiveIntegerField(default=1, null=True),
        ),
        migrations.AlterField(
            model_name='donationproject',
            name='contact_information',
            field=ckeditor.fields.RichTextField(help_text='Please provide the contact information of the person responsible for the conduction of this study. The contact information will be accessible for participants during the data donation.', null=True, verbose_name='Contact information'),
        ),
        migrations.AlterField(
            model_name='donationproject',
            name='data_protection_statement',
            field=ckeditor.fields.RichTextField(help_text='Please provide a data protection statement for your data donation collection. This should include the purpose for which the data is collected, how it will be stored, and who will have access to the data.', null=True, verbose_name='Data Protection Statement'),
        ),
        migrations.AlterField(
            model_name='questionbase',
            name='variable_name',
            field=models.SlugField(help_text='Will be used in the data export to identify responses to this question.', verbose_name='Variable name for storing response'),
        ),
    ]



import json

from bokeh.plotting import figure
from bokeh.embed import components

from django import template
from django.utils.safestring import mark_safe


register = template.Library()

figure_defaults = dict(
    toolbar_location='below',
    height=400,
    width=400
)

bar_plot_defaults = dict(
    color='navy',
    alpha=0.6,
    width=0.9
)


@register.simple_tag(takes_context=True)
def get_simple_bar_plot(context, x_data, bar_labels, figure_settings='{}', plot_settings='{}'):
    view = context.get('view', None)
    if view is None:
        return None

    figure_settings = json.loads(figure_settings)
    figure_settings = {**figure_defaults, **figure_settings}
    figure_settings['x_range'] = bar_labels

    plot_settings = json.loads(plot_settings)
    plot_settings = {**bar_plot_defaults, **plot_settings}

    bar_counts = [int(x) for x in x_data]

    try:
        plot = figure(**figure_settings)
        plot.vbar(x=bar_labels, top=bar_counts, **plot_settings)
        plot.toolbar.logo = None

        script, div = components(plot)
        view.extra_scripts.append(script)
        return mark_safe(div)

    except Exception as e:
        print(e)
        return None



import json
import zipfile

from django.http import HttpResponseRedirect
from django.shortcuts import redirect, reverse
from django.template import Context, Template
from django.utils import timezone
from django.utils.datastructures import MultiValueDictKeyError
from django.utils.decorators import method_decorator
from django.utils.safestring import SafeString
from django.views.generic.detail import DetailView
from django.views.decorators.cache import cache_page

from ddm.models.core import (
    DonationBlueprint, DonationProject, Participant, QuestionnaireResponse,
    FileUploader
)
from ddm.models.logs import ExceptionLogEntry, ExceptionRaisers
from ddm.models.questions import QuestionBase


class ParticipationFlowBaseView(DetailView):
    """
    Base class for participation flow views that implements base get and post
    methods, redirects and session management.
    """
    model = DonationProject
    context_object_name = 'project'

    steps = [
        'briefing',
        'data-donation',
        'questionnaire',
        'debriefing'
    ]
    participant = None
    current_step = None
    step_name = None

    def setup(self, request, *args, **kwargs):
        """ Initialize attributes shared by all participation steps. """
        super().setup(request, *args, **kwargs)
        self._initialize_values(request)

    def get(self, request, *args, **kwargs):
        # Redirect to previous step if necessary.
        if not self.steps[self.current_step] == self.step_name:
            return redirect(self.steps[self.current_step], slug=self.object.slug)

        # Render current view.
        context = self.get_context_data(object=self.object)
        self.extra_before_render(request)
        return self.render_to_response(context)

    def post(self, request, *arges, **kwargs):
        self.set_step_completed(request)
        return redirect(self.steps[self.current_step + 1],
                        slug=self.object.slug)

    def register_project_in_session(self, request):
        if not request.session.get(self.get_session_id()):
            request.session[self.get_session_id()] = {
                'participant_id': None,
                'last_completed_step': None
            }
            request.session.modified = True
        return

    def get_session_id(self):
        """
        Return id under which project related information is stored in a
        participant's session.
        """
        return f'project-{self.object.pk}'

    def get_participant_from_session(self, request):
        """
        Gets participant from session. If participant has not yet been created,
        creates new participant and saves it to session.
        """
        participant_id = request.session[self.get_session_id()]['participant_id']
        try:
            participant = Participant.objects.get(pk=participant_id)
        except Participant.DoesNotExist:
            participant = Participant.objects.create(
                project=self.object,
                start_time=timezone.now()
            )
            request.session[self.get_session_id()]['participant_id'] = participant.id
            request.session.modified = True
        return participant

    def get_current_step_from_session(self, request):
        """
        Gets current step from information stored in participant's session.
        """
        step = request.session[self.get_session_id()]['last_completed_step']
        if step is None:
            current_step = 0
        else:
            current_step = step + 1
        return current_step

    def set_step_completed(self, request):
        """
        Updates the last_completed_step attribute in current session.
        """
        request.session[self.get_session_id()]['last_completed_step'] = self.current_step
        request.session.modified = True
        return

    def extra_before_render(self, request):
        """
        Placeholder for stage specific code executed before rendering the
        response.
        """
        pass

    def _initialize_values(self, request):
        self.object = self.get_object()
        self.register_project_in_session(request)
        self.participant = self.get_participant_from_session(request)
        self.current_step = self.get_current_step_from_session(request)
        return


class BriefingView(ParticipationFlowBaseView):
    template_name = 'ddm/public/briefing.html'
    step_name = 'briefing'

    def post(self, request, *args, **kwargs):
        """
        Checks whether participant has provided briefing consent to continue with
        the study. If briefing consent is not given, redirects to end page.
        If briefing consent is not within the expected values, the briefing page
        is again returned with a form error.
        """
        # Check briefing consent if enabled for project.
        if self.get_object().briefing_consent_enabled:
            return self.check_consent(request, **kwargs)

        return super().post(request, **kwargs)

    def check_consent(self, request, **kwargs):
        """
        Checks whether post data contains information on briefing consent.
        Renders briefing view with error message if consent information is invalid.
        Renders debriefing view if no consent has been given.
        Renders next step if consent has been given.
        """
        # Check that answer has been provided and is valid.
        consent = request.POST.get('briefing_consent', None)
        if consent not in ['0', '1']:
            # Render briefing view again with error message.
            context = self.get_context_data(object=self.object)
            context.update({'briefing_error': True})
            return self.render_to_response(context)

        # Save consent to participant data.
        self.participant.extra_data['briefing_consent'] = consent
        self.participant.save()

        if consent == '0':
            # Redirect to debriefing page.
            request.session[self.get_session_id()]['last_completed_step'] = len(self.steps) - 2
            request.session.modified = True
            return redirect(self.steps[-1], slug=self.object.slug)
        else:
            return super().post(request, **kwargs)

    def extra_before_render(self, request):
        """ Extract URL parameters if this option has been enabled. """
        if self.object.url_parameter_enabled:
            # Only extract URL parameters on the first call of the view.
            if not self.participant.extra_data['url_param']:
                for param in self.object.get_expected_url_parameters():
                    self.participant.extra_data['url_param'][param] = request.GET.get(param, None)
                self.participant.save()
        return


@method_decorator(cache_page(0), name='dispatch')
class DataDonationView(ParticipationFlowBaseView):
    template_name = 'ddm/public/data_donation.html'
    step_name = 'data-donation'

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['uploader_configs'] = SafeString(self.get_uploader_configs())
        context['project_id'] = self.object.pk
        return context

    def get_uploader_configs(self):
        project_uploaders = FileUploader.objects.filter(project=self.object)
        uploader_configs = [fu.get_configs() for fu in project_uploaders]
        return json.dumps(uploader_configs)

    def post(self, request, *args, **kwargs):
        super().post(request, **kwargs)
        self.process_uploads(request.FILES)
        redirect_url = reverse(self.steps[self.current_step + 1],
                               kwargs={'slug': self.object.slug})
        return HttpResponseRedirect(redirect_url)

    def process_uploads(self, files):
        try:
            file = files['post_data']
        except MultiValueDictKeyError as err:
            msg = ('Data Donation Processing Exception: Did not receive '
                   f'expected data file from client. {err}')
            ExceptionLogEntry.objects.create(
                project=self.object,
                raised_by=ExceptionRaisers.SERVER,
                message=msg
            )
            return

        if not zipfile.is_zipfile(file):
            msg = ('Data Donation Processing Exception: Data file received '
                   'from client is not a zip file.')
            ExceptionLogEntry.objects.create(
                project=self.object,
                raised_by=ExceptionRaisers.SERVER,
                message=msg
            )
            return

        # Check if zip file contains expected file.
        unzipped_file = zipfile.ZipFile(file, 'r')
        if 'ul_data.json' not in unzipped_file.namelist():
            msg = ('Data Donation Processing Exception: "ul_data.json" is not '
                   'in namelist.')
            ExceptionLogEntry.objects.create(
                project=self.object,
                raised_by=ExceptionRaisers.SERVER,
                message=msg
            )
            return

        # Process donation data.
        file_data = json.loads(unzipped_file.read('ul_data.json').decode('utf-8'))
        for upload in file_data.keys():
            blueprint_id = upload
            blueprint_data = file_data[upload]
            try:
                blueprint = DonationBlueprint.objects.get(pk=blueprint_id,
                                                          project=self.object)
            except DonationBlueprint.DoesNotExist as e:
                msg = ('Data Donation Processing Exception: Referenced '
                       f'blueprint with id={blueprint_id} does not exist for '
                       'this project.')
                ExceptionLogEntry.objects.create(
                    project=self.object,
                    raised_by=ExceptionRaisers.SERVER,
                    message=msg
                )
                return
            blueprint.process_donation(blueprint_data, self.participant)


class QuestionnaireView(ParticipationFlowBaseView):
    template_name = 'ddm/public/questionnaire.html'
    step_name = 'questionnaire'

    def setup(self, request, *args, **kwargs):
        """ Reset extra_scripts """
        super().setup(request, *args, **kwargs)
        self.extra_scripts = []

    def get(self, request, *args, **kwargs):
        """
        Skip questionnaire if no questions are defined and redirect to next step.
        Otherwise, render questionnaire.
        """
        # Redirect to previous step if necessary.
        if not self.steps[self.current_step] == self.step_name:
            return redirect(self.steps[self.current_step], slug=self.object.slug)

        context = self.get_context_data(object=self.object)
        if not len(context['q_config']) > 2:
            self.set_step_completed(request)
            return redirect(self.steps[self.current_step + 1],
                            slug=self.object.slug)
        else:
            return self.render_to_response(context)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        question_config = self.object.get_questionnaire_config(self.participant, self)
        context['q_config'] = json.dumps(question_config)
        context['extra_scripts'] = set(self.extra_scripts)
        return context

    def post(self, request, *args, **kwargs):
        super().post(request, **kwargs)
        self.process_response(request.POST)
        return redirect(self.steps[self.current_step + 1],
                        slug=self.object.slug)

    def process_response(self, response):
        try:
            post_data = json.loads(response['post_data'])
        except MultiValueDictKeyError:
            msg = ('Questionnaire Post Exception: POST did not contain '
                   'expected key "post_data".')
            ExceptionLogEntry.objects.create(
                project=self.object,
                raised_by=ExceptionRaisers.SERVER,
                message=msg
            )
            return

        for question_id in post_data:
            try:
                question = QuestionBase.objects.get(pk=int(question_id))
            except QuestionBase.doesNotExist:
                msg = ('Questionnaire Post Exception:'
                       f'Question with id={question_id} does not exist.')
                ExceptionLogEntry.objects.create(
                    project=self.object,
                    raised_by=ExceptionRaisers.SERVER,
                    message=msg
                )
                continue
            except ValueError:
                msg = ('Questionnaire Post Exception: Received invalid '
                       f'question_id ({question_id}) in questionnaire '
                       f'post_data.')
                ExceptionLogEntry.objects.create(
                    project=self.object,
                    raised_by=ExceptionRaisers.SERVER,
                    message=msg
                )
                continue
            question.validate_response(post_data[question_id])

        self.save_response(response)

    def save_response(self, response):
        QuestionnaireResponse.objects.create(
            project=self.object,
            participant=self.participant,
            time_submitted=timezone.now(),
            data=response
        )


class DebriefingView(ParticipationFlowBaseView):
    template_name = 'ddm/public/debriefing.html'
    step_name = 'debriefing'

    def get_context_data(self, **kwargs):
        """ Inject url parameters in redirect target. """
        context = super().get_context_data(**kwargs)
        if self.object.redirect_enabled:
            template = Template(self.object.redirect_target)
            template_context = self.participant.extra_data['url_param']
            template_context['ddm_participant_id'] = self.participant.pk
            template_context['ddm_project_id'] = self.object.pk
            context['redirect_target'] = template.render(Context(template_context))
        else:
            context['redirect_target'] = None
        return context

    def extra_before_render(self, request):
        """Set step to completed and update participant information."""
        if not self.participant.completed:
            self.participant.end_time = timezone.now()
            self.participant.completed = True
            self.participant.save()
        return



from django.test import SimpleTestCase
from django.urls import reverse, resolve
from ddm.views import participation_flow


class TestParticipationFlowUrls(SimpleTestCase):

    def test_project_briefing_url_resolves(self):
        url = reverse('briefing', args=['project-slug'])
        self.assertEqual(
            resolve(url).func.view_class,
            participation_flow.BriefingView
        )

    def test_project_datadonation_url_resolves(self):
        url = reverse('data-donation', args=['project-slug'])
        self.assertEqual(
            resolve(url).func.view_class,
            participation_flow.DataDonationView
        )

    def test_project_questionnaire_url_resolves(self):
        url = reverse('questionnaire', args=['project-slug'])
        self.assertEqual(
            resolve(url).func.view_class,
            participation_flow.QuestionnaireView
        )

    def test_project_debriefing_url_resolves(self):
        url = reverse('debriefing', args=['project-slug'])
        self.assertEqual(
            resolve(url).func.view_class,
            participation_flow.DebriefingView
        )



from django.contrib.auth import get_user_model
from django.test import TestCase, override_settings
from django.utils import timezone

from ddm.models.core import (
    DataDonation, DonationProject, QuestionnaireResponse, ResearchProfile,
    DonationBlueprint, Participant
)
from ddm.models.encryption import Encryption, Decryption


User = get_user_model()


class TestEncryptedFieldIsolated(TestCase):
    def test_encrypt_decrypt(self):
        text = 'test_string'
        enc = Encryption(secret='foo', salt='bar')
        enc_text = enc.encrypt(text)
        dec = Decryption(secret='foo', salt='bar')
        self.assertNotEqual(text, enc_text)
        self.assertEqual(text, dec.decrypt(enc_text))

    def test_encrypt_decrypt_explicit_public(self):
        text = 'test_string'
        dec = Decryption(secret='foo', salt='bar')
        pub = Encryption.get_public_key(secret='foo', salt='bar')
        enc_text = Encryption(public=pub).encrypt(text)
        self.assertNotEqual(text, enc_text)
        self.assertEqual(text, dec.decrypt(enc_text))


@override_settings(DDM_SETTINGS={'EMAIL_PERMISSION_CHECK':  r'.*(\.|@)mail\.com$', })
class TestModelEncryption(TestCase):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()

        base_user = User.objects.create_user(**{
            'username': 'owner', 'password': '123', 'email': 'owner@mail.com'
        })
        base_profile = ResearchProfile.objects.create(user=base_user)

        cls.base_project = DonationProject.objects.create(
            name='Base Project', slug='base', owner=base_profile)

        cls.alt_project = DonationProject.objects.create(
            name='Test Project Custom',
            slug='test-project-custom',
            super_secret=True,
            secret_key='test1234',
            owner=base_profile
        )

        cls.base_blueprint = DonationBlueprint.objects.create(
            project=cls.base_project,
            name='donation blueprint',
            expected_fields='"a", "b"',
            file_uploader=None
        )
        cls.alt_blueprint = DonationBlueprint.objects.create(
            project=cls.alt_project,
            name='donation blueprint',
            expected_fields='"a", "b"',
            file_uploader=None
        )

        cls.base_participant = Participant.objects.create(
            project=cls.base_project,
            start_time=timezone.now()
        )
        cls.alt_participant = Participant.objects.create(
            project=cls.alt_project,
            start_time=timezone.now()
        )

        cls.raw_data_short = '{"some_data": "some_value"}'
        cls.raw_data_long = '{' + 100*'"some_data": "some_value",' + '}'

    def test_data_donation_encryption_default(self):
        for raw_data in [self.raw_data_short, self.raw_data_long]:
            with self.subTest(raw_data=raw_data):
                dd = DataDonation.objects.create(
                    project=self.base_project,
                    blueprint=self.base_blueprint,
                    participant=self.base_participant,
                    consent=True,
                    status='some status',
                    data=raw_data
                )
                self.assertNotEqual(raw_data, dd.data)
                self.assertEqual(raw_data, dd.get_decrypted_data())

    def test_questionnaire_response_encryption_default(self):
        for raw_data in [self.raw_data_short, self.raw_data_long]:
            with self.subTest(raw_data=raw_data):
                qr = QuestionnaireResponse.objects.create(
                    project=self.base_project,
                    participant=self.base_participant,
                    data=raw_data
                )
                self.assertNotEqual(raw_data, qr.data)
                self.assertEqual(raw_data, qr.get_decrypted_data())

    def test_data_donation_encryption_super_secret(self):
        for raw_data in [self.raw_data_short, self.raw_data_long]:
            with self.subTest(raw_data=raw_data):
                dd = DataDonation.objects.create(
                    project=self.alt_project,
                    blueprint=self.alt_blueprint,
                    participant=self.alt_participant,
                    consent=True,
                    status='some status',
                    data=raw_data,
                )
                self.assertNotEqual(raw_data, dd.data)
                with self.assertRaises(KeyError):
                    dd.get_decrypted_data()
                with self.assertRaises(ValueError):
                    dd.get_decrypted_data(secret='invalidSecret1234')
                self.assertEqual(raw_data, dd.get_decrypted_data(secret='test1234'))

    def test_questionnaire_response_encryption_super_secret(self):
        for raw_data in [self.raw_data_short, self.raw_data_long]:
            with self.subTest(raw_data=raw_data):
                qr = QuestionnaireResponse.objects.create(
                    project=self.alt_project,
                    participant=self.alt_participant,
                    data=raw_data
                )
                self.assertNotEqual(raw_data, qr.data)
                with self.assertRaises(KeyError):
                    qr.get_decrypted_data()
                with self.assertRaises(ValueError):
                    qr.get_decrypted_data(secret='invalidSecret1234')
                self.assertEqual(raw_data, qr.get_decrypted_data(secret='test1234'))



from django.contrib.auth import get_user_model
from django.test import TestCase

from ddm.models.core import (
    DonationInstruction, ResearchProfile, DonationProject, FileUploader
)


User = get_user_model()


class TestDonationInstructionModel(TestCase):
    @classmethod
    def setUpTestData(cls):
        user = User.objects.create_user(**{
            'username': 'owner', 'password': '123', 'email': 'owner@mail.com'
        })
        profile = ResearchProfile.objects.create(user=user)

        project = DonationProject.objects.create(
            name='Base Project', slug='base', owner=profile)

        cls.file_uploader = FileUploader.objects.create(
            project=project,
            name='basic file uploader',
            upload_type=FileUploader.UploadTypes.SINGLE_FILE
        )

    def setUp(self):
        self.inst_2 = DonationInstruction.objects.create(
            text='instruction 2',
            index=2,
            file_uploader=self.file_uploader
        )
        self.inst_3 = DonationInstruction.objects.create(
            text='instruction 3',
            index=3,
            file_uploader=self.file_uploader
        )

    def test_create_with_existing_index_pushes_other_indices(self):
        new_instruction = DonationInstruction.objects.create(
            text='instruction new',
            index=2,
            file_uploader=self.file_uploader
        )
        self.assertEqual(new_instruction.index, 2)

        self.inst_2.refresh_from_db()
        self.assertEqual(self.inst_2.index, 3)

        self.inst_3.refresh_from_db()
        self.assertEqual(self.inst_3.index, 4)

    def test_decrease_index_of_existing_instruction(self):
        self.inst_3.index = 2
        self.inst_3.save()
        self.assertEqual(self.inst_3.index, 2)

        self.inst_2.refresh_from_db()
        self.assertEqual(self.inst_2.index, 3)

    def test_increase_index_of_existing_instruction(self):
        self.inst_2.index = 3
        self.inst_2.save()
        self.assertEqual(self.inst_2.index, 3)

        self.inst_3.refresh_from_db()
        self.assertEqual(self.inst_3.index, 2)

    def test_indices_are_adjusted_on_delete(self):
        self.inst_2.delete()

        self.inst_3.refresh_from_db()
        self.assertEqual(self.inst_3.index, 2)



from django.contrib import messages
from django.contrib.messages.views import SuccessMessageMixin
from django.urls import reverse_lazy, reverse
from django.views.generic import TemplateView
from django.views.generic.detail import DetailView
from django.views.generic.edit import CreateView, UpdateView, DeleteView, FormView
from django.views.generic.list import ListView

from ddm.forms import APITokenCreationForm, ProjectCreateForm
from ddm.models.auth import ProjectAccessToken
from ddm.models.core import DonationProject, ResearchProfile
from ddm.views.admin.auth import DdmAuthMixin


class ProjectList(DdmAuthMixin, ListView):
    """ View to display a list of existing donation projects. """
    model = DonationProject
    template_name = 'ddm/admin/project/list.html'

    def get_queryset(self):
        return DonationProject.objects.filter(owner__user=self.request.user)


class ProjectCreate(SuccessMessageMixin, DdmAuthMixin, CreateView):
    """ View to create a new donation project. """
    model = DonationProject
    template_name = 'ddm/admin/project/create.html'
    form_class = ProjectCreateForm
    success_message = 'Project was created successfully.'

    def get_initial(self):
        self.initial = super().get_initial()
        self.initial.update({'owner': ResearchProfile.objects.get(user=self.request.user)})
        return self.initial

    def form_valid(self, form):
        form.instance.owner = ResearchProfile.objects.get(user=self.request.user)
        return super().form_valid(form)


class ProjectDetail(DdmAuthMixin, DetailView):
    """ View to display landing page for project. """
    model = DonationProject
    template_name = 'ddm/admin/project/detail.html'


class ProjectEdit(SuccessMessageMixin, DdmAuthMixin, UpdateView):
    """ View to edit the details of an existing donation project. """
    model = DonationProject
    template_name = 'ddm/admin/project/edit.html'
    fields = [
        'name', 'slug', 'contact_information', 'data_protection_statement',
        'url_parameter_enabled', 'expected_url_parameters',
        'redirect_enabled', 'redirect_target',
        'img_header_left', 'img_header_right'
    ]
    success_message = 'Project details successfully updated.'

    def form_valid(self, form):
        if form.cleaned_data['url_parameter_enabled'] and form.cleaned_data['expected_url_parameters'] == '':
            form.add_error('expected_url_parameters', 'URL parameter is enabled but no parameter is defined.')

        if form.cleaned_data['redirect_enabled'] and form.cleaned_data['redirect_target'] == '':
            form.add_error('redirect_target', 'Redirect is enabled but no redirect target is defined.')

        if form.is_valid():
            return super().form_valid(form)
        else:
            return self.form_invalid(form)


class ProjectDelete(DdmAuthMixin, DeleteView):
    """ View to display a list of existing donation projects. """
    model = DonationProject
    template_name = 'ddm/admin/project/delete.html'
    success_url = reverse_lazy('project-list')
    success_message = 'Project was deleted.'

    def delete(self, request, *args, **kwargs):
        messages.success(self.request, self.success_message)
        return super().delete(request, *args, **kwargs)


class BriefingEdit(SuccessMessageMixin, DdmAuthMixin, UpdateView):
    """ View to edit the briefing page. """
    model = DonationProject
    template_name = 'ddm/admin/project/edit-briefing.html'
    fields = ['briefing_text', 'briefing_consent_enabled', 'briefing_consent_label_yes', 'briefing_consent_label_no']
    success_message = 'Briefing page successfully updated.'

    def form_valid(self, form):
        consent_label_error_msg = 'When briefing consent is enabled, a consent label must be provided.'
        if form.cleaned_data['briefing_consent_enabled'] and form.cleaned_data['briefing_consent_label_yes'] == '':
            form.add_error('briefing_consent_label_yes', consent_label_error_msg)

        if form.cleaned_data['briefing_consent_enabled'] and form.cleaned_data['briefing_consent_label_no'] == '':
            form.add_error('briefing_consent_label_no', consent_label_error_msg)

        if form.is_valid():
            return super().form_valid(form)
        else:
            return self.form_invalid(form)


class DebriefingEdit(SuccessMessageMixin, DdmAuthMixin, UpdateView):
    """ View to edit the debriefing page. """
    model = DonationProject
    template_name = 'ddm/admin/project/edit-debriefing.html'
    fields = ['debriefing_text']
    success_message = 'Debriefing page successfully updated.'


class ProjectLogsView(SuccessMessageMixin, DdmAuthMixin, TemplateView):
    """ View that lists all exceptions related to a project. """
    template_name = 'ddm/admin/project/logs/overview.html'

    def get_project(self):
        project_id = self.kwargs.get('project_pk')
        return DonationProject.objects.filter(pk=project_id).first()

    def get_event_logs(self):
        project = self.get_project()
        return project.eventlogentry_set.all()

    def get_exception_logs(self):
        project = self.get_project()
        return project.exceptionlogentry_set.all()

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context.update({
            'project': self.get_project(),
            'events': self.get_event_logs(),
            'exceptions': self.get_exception_logs()
        })
        return context


class ProjectAPITokenView(SuccessMessageMixin, DdmAuthMixin, FormView):
    """ View to see existing access token or generate a new one. """
    template_name = 'ddm/admin/project/token.html'
    form_class = APITokenCreationForm

    def get_project(self):
        """ Returns current project. """
        project_id = self.kwargs.get('pk')
        return DonationProject.objects.filter(pk=project_id).first()

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        project = self.get_project()
        context.update({
            'project': project,
            'token': ProjectAccessToken.objects.filter(project=project).first()
        })
        return context

    def form_valid(self, form):
        """ If form is valid, either create new token or delete existing one. """
        project = self.get_project()
        if form.cleaned_data['action'] == 'create':
            project.create_token(expiration_days=form.cleaned_data['expiration_days'])

        if form.cleaned_data['action'] == 'delete':
            token = project.get_token()
            token.delete()

        return super().form_valid(form)

    def get_success_url(self):
        return reverse('project-token', kwargs={'pk': self.kwargs.get('pk')})



# Generated by Django 3.2.13 on 2023-01-05 15:02

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('ddm', '0024_customtoken'),
    ]

    operations = [
        migrations.RenameModel(
            old_name='CustomToken',
            new_name='ProjectAccessToken',
        ),
    ]



# Generated by Django 3.2.13 on 2022-06-29 14:51

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ddm', '0009_auto_20220629_1641'),
    ]

    operations = [
        migrations.AlterField(
            model_name='donationblueprint',
            name='expected_fields',
            field=models.TextField(validators=[django.core.validators.RegexValidator('^((["][^"]+["]))(\\s*,\\s*((["][^"]+["])))*[,\\s]*$', message='Field must contain one or multiple comma separated strings. Strings must be enclosed in double quotes ("string").')]),
        ),
        migrations.AlterField(
            model_name='donationblueprint',
            name='extracted_fields',
            field=models.TextField(blank=True, null=True, validators=[django.core.validators.RegexValidator('^((["][^"]+["]))(\\s*,\\s*((["][^"]+["])))*[,\\s]*$', message='Field must contain one or multiple comma separated strings. Strings must be enclosed in double quotes ("string").')]),
        ),
    ]



# Generated by Django 3.2.13 on 2022-09-28 11:44

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('ddm', '0015_auto_20220923_1055'),
    ]

    operations = [
        migrations.AlterField(
            model_name='donationproject',
            name='redirect_target',
            field=models.CharField(blank=True, help_text='Always include <i>http://</i> or <i>https://</i> in the redirect target. If URL parameter extraction is enabled for this project, you can include the extracted URL parameters in the redirect target as follows: "https://redirect.me/?redirectpara=<b>{{URLParameter}}</b>".', max_length=2000, null=True, verbose_name='Redirect target'),
        ),
        migrations.CreateModel(
            name='ProcessingRule',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=250)),
                ('field', models.TextField()),
                ('execution_order', models.IntegerField()),
                ('input_type', models.CharField(choices=[('number', 'Number'), ('date', 'Date'), ('string', 'String')], default='string', max_length=10)),
                ('comparison_operator', models.CharField(choices=[('==', 'Equal (==)'), ('!=', 'Not Equal (!=)'), ('>', 'Greater than (>)'), ('<', 'Smaller than (<)'), ('>=', 'Greater than or equal (>=)'), ('<=', 'Smaller than or equal (<=)'), ('regex', 'Regex (removes matches)')], default='==', max_length=10)),
                ('comparison_value', models.TextField()),
                ('blueprint', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='ddm.donationblueprint')),
            ],
        ),
    ]
